/*
   XGDragView - Drag and Drop code for X11 backends.

   Copyright (C) 1998,1999,2001 Free Software Foundation, Inc.

   Created by: Wim Oudshoorn <woudshoo@xs4all.nl>
   Date: Nov 2001

   Written by:  Adam Fedor <fedor@gnu.org>
   Date: Nov 1998
   
   This file is part of the GNU Objective C User Interface Library.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   */


#include <AppKit/NSView.h>

#include "SharedX/XGDragView.h"

#ifdef XDPS_BACKEND_LIBRARY
#define XGContext       NSDPSContext
#define XGDrawingEngine PXKDrawingEngine
#define XGDragView	PXKDragView
#include "gnustep/xdps/NSDPSContextWindow.h"
#include "PXKDrawingEngine.h"
#else
#include "gnustep/xgps/XGContext.h"
#include "gnustep/xgps/XGContextPrivate.h"
#include "gnustep/xgps/XGContextWindow.h"
#endif

#include <X11/extensions/shape.h>

/* Size of the dragged window */
#define	DWZ	48
#define ALPHA_THRESHOLD 158

#define XDPY  [XGContext currentXDisplay]

@interface XGRawWindow : NSWindow
@end

@interface NSImage (BackEnd)
- (Pixmap) xPixmapMask;
@end

@interface NSCursor (BackendPrivate)
- (void *)_cid;
@end

// --- DRAG AND DROP SUPPORT  (XDND) -----------------------------------

/*
 * This will get initialized when we declare a window that will
 * accept dragging or if we start a dragging ourself. Up to than
 * even the dragging messages are not defined.
 */
static DndClass dnd;
static BOOL     xDndInitialized = NO;

void
GSEnsureDndIsInitialized (void)
{
  if (xDndInitialized == NO)
    {
      xDndInitialized = YES;
      xdnd_init (&dnd, XDPY);
    }
}


DndClass xdnd (void)
{
  return dnd;			// FIX ME rename with private desig
}


Atom 
GSActionForDragOperation(unsigned int op)
{
  Atom xaction;
  if (op == NSDragOperationAll)
    xaction = dnd.XdndActionPrivate;
  else if (op & NSDragOperationCopy)
    xaction = dnd.XdndActionCopy;
  else if (op & NSDragOperationLink)
    xaction = dnd.XdndActionLink;
  else if (op & NSDragOperationGeneric)
    xaction = dnd.XdndActionCopy;
  else if (op & NSDragOperationPrivate)
    xaction = dnd.XdndActionPrivate;
  else 
    xaction = None;
  return xaction;
}


NSDragOperation
GSDragOperationForAction(Atom xaction)
{
  NSDragOperation action;
  if (xaction == dnd.XdndActionCopy)
    action = NSDragOperationCopy;
  else if (xaction == dnd.XdndActionMove)
    action = NSDragOperationCopy;
  else if (xaction == dnd.XdndActionLink) 
    action = NSDragOperationLink;
  else if (xaction == dnd.XdndActionAsk) 
    action = NSDragOperationGeneric;
  else if (xaction == dnd.XdndActionPrivate) 
    action = NSDragOperationPrivate;
  else
    action = NSDragOperationNone;
  return action;
}



static inline
Atom *
mimeTypeForPasteboardType(Display *xDisplay, NSZone *zone, NSArray *types)
{
  Atom	*typelist;
  int	count = [types count];
  int	i;

  typelist = NSZoneMalloc(zone, (count+1) * sizeof(Atom));
  for (i = 0; i < count; i++)
    {
      NSString	*mime;

      mime = [types objectAtIndex: i];
      mime = [NSPasteboard mimeTypeForPasteboardType: mime];
      typelist[i] = XInternAtom(xDisplay, [mime cString], False);
    }
  typelist[count] = 0;

  return typelist;
}



@implementation XGDragView

static	XGDragView	*sharedDragView = nil;

+ (XGDragView*) sharedDragView
{
  if (sharedDragView == nil)
    {
      GSEnsureDndIsInitialized ();
      sharedDragView = [XGDragView new];
    }
  return sharedDragView;
}

- (id) init
{
  NSRect	winRect = {{0, 0}, {DWZ, DWZ}};
  XGRawWindow	*sharedDragWindow = [XGRawWindow alloc];

  self = [super init];
  if (self != nil)
    {
      dragCell = [[NSCell alloc] initImageCell: nil];
      [dragCell setBordered: NO];
      
      [sharedDragWindow initWithContentRect: winRect
				  styleMask: NSBorderlessWindowMask
				    backing: NSBackingStoreNonretained
				      defer: YES];

      [sharedDragWindow setContentView: self];

      RELEASE(self);
    }
  return self;
}

- (void) drawRect: (NSRect)rect
{
  [dragCell drawWithFrame: rect inView: self];
}


/*
 * External drag operation
 */
- (void) setupDragInfoFromXEvent: (XEvent *)xEvent
{
  // Start a dragging session from another application
  dragSource = nil;
  dragExternal = YES;
  ASSIGN(dragPasteboard, [NSPasteboard pasteboardWithName: NSDragPboard]);
}

- (void) updateDragInfoFromEvent: (NSEvent*)event
{
  // Store the drag info, so that we can send status messages as response 
  dragWindow = [event window];
  dragPoint = [event locationInWindow];
  dragSequence = [event timestamp];
  dragMask = [event data2];
}

- (void) resetDragInfo
{
  DESTROY(dragPasteboard);
}

/*
 * Local drag operation
 */
- (void) dragImage: (NSImage*)anImage
		at: (NSPoint)viewLocation
	    offset: (NSSize)initialOffset
	     event: (NSEvent*)event
	pasteboard: (NSPasteboard*)pboard
	    source: (id)sourceObject
	 slideBack: (BOOL)slideFlag
{
  if (anImage == nil)
    {
      anImage = [NSImage imageNamed: @"common_Close"];
    }

  [dragCell setImage: anImage];

  ASSIGN(dragPasteboard, pboard);
  dragSource = sourceObject;
  dragSequence = [event timestamp];
  dragExternal = NO;

  NSDebugLLog(@"NSDragging", @"Start drag with %@", [pboard types]);
  typelist = mimeTypeForPasteboardType (XDPY, [self zone], [pboard types]);
  [self _handleDrag: event];
  NSZoneFree([self zone], typelist);
  typelist = NULL;
}

- (void) _sendLocalEvent: (GSAppKitSubtype)subtype
		  action: (NSDragOperation)action
	        position: (NSPoint)eventLocation
	       timestamp: (NSTimeInterval)time
	        toWindow: (NSWindow*)dWindow
{
  NSEvent *e;
  NSGraphicsContext *context = GSCurrentContext();
  gswindow_device_t *windev;

  windev = [XGContext _windowWithTag: [dWindow windowNumber]];
  eventLocation = NSMakePoint(eventLocation.x - NSMinX(windev->xframe),
			      eventLocation.y - NSMinY(windev->xframe));
  eventLocation.y = NSHeight(windev->xframe) - eventLocation.y;

  e = [NSEvent otherEventWithType: NSAppKitDefined
	  location: eventLocation
	  modifierFlags: 0
	  timestamp: time
	  windowNumber: windev->number
	  context: context
	  subtype: subtype
	  data1: windev->ident
	  data2: action];
  [dWindow sendEvent: e];
}

/*
  Method to initialize the dragview before it is put on the screen.
  It only initializes the instance variables that have to do with
  moving the image over the screen and variables that are used
  to keep track where we.

  So it is typically used just before the dragview is actually displayed.

  Pre coniditions:
  - dragCell is initialized with the image to drag.
  - typelist is initialized with the dragging types
  Post conditions:
  - all instance variables pertaining to moving the window are initialized
  - all instance variables pertaining to X-Windows  are initialized
  
 */
- (void) _setupWindow: (NSPoint) dragStart
{
  NSSize		imageSize = [[dragCell image] size];
  Pixmap		pixmap = 0;
  
  offset = NSMakePoint (imageSize.width / 2.0, imageSize.height / 2.0);
  
  [_window setFrame: NSMakeRect (dragStart.x - offset.x, 
                                 dragStart.y - offset.y,
                                 imageSize.width, imageSize.height)
           display: NO];
  
  /* Hack? Need to make sure the window backend exists before we try
     X calls */
  if ([_window gState] == 0)
    {
      [_window orderFront: nil];
    }

  dragWindev   = [XGContext _windowWithTag: [_window windowNumber]];
  targetWindow = 0;

  NSDebugLLog (@"NSDragging", @"---dragWindow: %x <- %x",
                 dragWindev->parent, dragWindev->ident);

  /* setup the wx and wy coordinates, used for moving the view around */
  wx = dragWindev->siz_hints.x;
  wy = dragWindev->siz_hints.y;

  dragPosition = dragStart;
  newPosition = dragStart;

  // Register our window for the current DnD types
  // CHECKME: Should this THIS window or the parent?
  xdnd_set_dnd_aware(&dnd, dragWindev->ident, typelist);

  // FIXME this should be integrated with xPixmapMask
  // but that method should be reorganized
  if ([[[dragCell image] backgroundColor] alphaComponent] * 256 <= ALPHA_THRESHOLD)
    {
      // Make it a shaped window containing a pixmap
      // Need to lockFocus to do this. FIXME when image caching works
      [self lockFocus];
      pixmap = [[dragCell image] xPixmapMask];
      [self unlockFocus];
    }
  
  if (pixmap)
    {
      XShapeCombineMask(XDPY, dragWindev->ident, ShapeBounding, 0, 0,
	pixmap, ShapeSet);
      XFreePixmap(XDPY, pixmap);
    }
  else
    {
      XShapeCombineMask(XDPY, dragWindev->ident, ShapeBounding, 0, 0,
	0, ShapeSet);
    }

  [_window orderFront: nil];

  
// NOTE: This should be done at the begin of the drag, but 
  // xpbs.m will already do this as the data gets into the pasteboard.
  // Get the Xdnd Selection.
  // xdnd_set_selection_owner(&dnd, windev->ident, None);
}


/*
  updates the operationMask by examining modifier keys
  pressed during -theEvent-.

  If the current value of operationMask == NSDragOperationIgnoresModifiers
  it will return immediately without updating the operationMask
  
  This method will return YES if the operationMask
  is changed, NO if it is still the same.
*/
  
- (BOOL) _updateOperationMask: (NSEvent*) theEvent
{
  unsigned int mod = [theEvent modifierFlags];
  unsigned int oldOperationMask = operationMask;

  if (operationMask == NSDragOperationIgnoresModifiers)
    {
      return NO;
    }
  
  if (mod & NSControlKeyMask)
    {
      operationMask = NSDragOperationLink;
    }
  else if (mod & NSAlternateKeyMask)
    {
      operationMask = NSDragOperationCopy;
    }
  else if (mod & NSCommandKeyMask)
    {
      operationMask = NSDragOperationGeneric;
    }
  else
    {
      operationMask = NSDragOperationAll;
    }

  return (operationMask != oldOperationMask);
}
  
/*
  _setCursor examines the state of the dragging and update
  the cursor accordingly.  It will save the current cursor,
  if you want to keep the original you have to save it yourself.

  The code recogines 4 cursors:

  - NONE - when the source does not allow dragging
  - COPY - when the current operation is ONLY Copy
  - LINK - when the current operation is ONLY Link
  - GENERIC - all other cases

  And two colors

  - GREEN - when the target accepts the drop
  - BLACK - when the target does not accept the drop

  Note that the code to figure out which of the 4 cursor to use
  depends on the fact that

  {NSDragOperationNone, NSDragOperationCopy, NSDragOperationLink} = {0, 1, 2}
*/
- (void) _setCursor
{
  NSCursor *newCursor = nil;
  NSImage  *image = nil;
  int       mask;

  mask = dragMask & operationMask;

  if (targetWindow)
    mask &= targetMask;

  NSDebugLLog (@"NSDragging",
               @"drag, operation, target mask = (%x, %x, %x), dnd aware = %d\n",
               dragMask, operationMask, targetMask,
               (targetWindow != (Window) None));
  
  
  switch (mask)
    {
    case NSDragOperationNone:
      image = [NSImage imageNamed: @"common_noCursor"];
      newCursor = [[NSCursor alloc] initWithImage: image];
      break;
    case NSDragOperationCopy:
      image = [NSImage imageNamed: @"common_copyCursor"];
      newCursor = [[NSCursor alloc] initWithImage: image];
      break;
    case NSDragOperationLink:
      image = [NSImage imageNamed: @"common_linkCursor"];
      newCursor = [[NSCursor alloc] initWithImage: image];
      break;
    case NSDragOperationGeneric:
      break;
    default:
      // FIXME: Should not happen, add warning?
    }

  if (newCursor == nil)
    {
      newCursor = [[NSCursor arrowCursor] retain];
    }
  
  [newCursor set];

  if ((targetWindow != (Window) None) && mask != NSDragOperationNone)
    {
      DPSsetcursorcolor(GSCurrentContext (), 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, [newCursor _cid]);
    }
  else
    {
      DPSsetcursorcolor(GSCurrentContext (), 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, [newCursor _cid]);
    }
  [newCursor release];
}

/*
  The dragging support works by hijacking the NSApp event loop.

  - this function loops until the dragging operation is finished
    and consumes all NSEvents during the drag operation.

  - It sets up periodic events.  The drawing and communication
    with DraggingSource and DraggingTarget is handled in the
    periodic event code.  The use of periodic events is purely
    a performance improvement.  If no periodic events are used
    the system can not process them all on time.
    At least on a 333Mhz laptop, using fairly simple
    DraggingTarget code.

  PROBLEMS:

  - No autoreleasePools are created.  So long drag operations can consume
    memory

  - It seems that sometimes a periodic event get lost.
*/
- (void) _handleDrag: (NSEvent*)theEvent
{
  // Caching some often used values. These values do not
  // change in this method.
  Display	*xDisplay = [XGContext currentXDisplay];
  NSWindow	*eWindow = [theEvent window]; // used for coordination transformation
  NSDate	*theDistantFuture = [NSDate distantFuture];
  NSImage       *dragImage = [dragCell image];
  unsigned int	eventMask = NSLeftMouseDownMask | NSLeftMouseUpMask
			    | NSLeftMouseDraggedMask | NSMouseMovedMask
			    | NSPeriodicMask | NSAppKitDefinedMask | NSFlagsChangedMask;

  NSPoint       point;
  
  // Storing values, to restore after we have finished.
  NSCursor      *cursorBeforeDrag = [NSCursor currentCursor];
  
  point = [eWindow convertBaseToScreen: [theEvent locationInWindow]];

  [self _setupWindow: point];

  // Notify the source that dragging has started
  if ([dragSource respondsToSelector:
      @selector(draggedImage:beganAt:)])
    {
      [dragSource draggedImage: dragImage
		  beganAt: point];
    }

  NSDebugLLog(@"NSDragging", @"Drag window X origin %d %d\n", wx, wy);

  
  // --- Setup up the masks for the drag operation ---------------------
  
  if ([dragSource respondsToSelector: @selector(ignoreModifierKeysWhileDragging)]
      && [dragSource ignoreModifierKeysWhileDragging])
    {
      operationMask = NSDragOperationIgnoresModifiers;
    }
  else
    {
      operationMask = 0;
      [self _updateOperationMask: theEvent];
    }

  dragMask = [dragSource draggingSourceOperationMaskForLocal: !dragExternal];
  targetMask = NSDragOperationAll;
  
  // --- Setup the event loop ------------------------------------------

  [self _updateAndMoveImageToCorrectPosition];
  [NSEvent startPeriodicEventsAfterDelay: 0.02 withPeriod: 0.03];

  // --- Loop that handles all events durring drag operation -----------
  
  while ([theEvent type] != NSLeftMouseUp)
    {
      [self _handleEventDuringDragging: theEvent];

      theEvent = [NSApp nextEventMatchingMask: eventMask
				    untilDate: theDistantFuture
				       inMode: NSEventTrackingRunLoopMode
				      dequeue: YES];
    }


  // --- Event loop for drag opeeration stopped ------------------------
  
  [NSEvent stopPeriodicEvents];
  [self _updateAndMoveImageToCorrectPosition];

  [_window orderOut: nil];

  NSDebugLLog(@"NSDragging", @"dnd ending %x\n", targetWindow);


  // --- Deposit the drop ----------------------------------------------
  
  if (targetWindow != (Window) None)
    {
      NSDebugLLog(@"NSDragging", @"sending dnd drop\n");
      if (!dragExternal)
	{
	  [self _sendLocalEvent: GSAppKitDraggingDrop
			 action: 0
		       position: NSZeroPoint
		      timestamp: CurrentTime
		       toWindow: dragWindow];
	}
      else
	{
	  xdnd_send_drop(&dnd, targetWindow, dragWindev->ident, CurrentTime);
	}

      //CHECKME: Why XSync here?
      XSync(xDisplay, False);
      if ([dragSource respondsToSelector:
	@selector(draggedImage:endedAt:deposited:)])
	{
	  point = [theEvent locationInWindow];
	  point = [[theEvent window] convertBaseToScreen: point];
	  [dragSource draggedImage: dragImage
			   endedAt: point
			 deposited: YES];
	}
    }
  else
    {
      if (targetWindow == dragWindev->root)
        {
	  // FIXME There is an xdnd extension for root drop
	}
      if ([dragSource respondsToSelector:
	@selector(draggedImage:endedAt:deposited:)])
	{
	  point = [theEvent locationInWindow];
	  point = [[theEvent window] convertBaseToScreen: point];
	  [dragSource draggedImage: dragImage
			   endedAt: point
			 deposited: NO];
	}
    }
  [cursorBeforeDrag set];
}


/*
 * Handle the events for the event loop during drag and drop
 */
- (void) _handleEventDuringDragging: (NSEvent *) theEvent
{
  switch ([theEvent type])
    {
    case  NSAppKitDefined:
      {
        GSAppKitSubtype	sub = [theEvent subtype];
        
        switch (sub)
        {
        case GSAppKitWindowMoved:
          /*
           * Keep window up-to-date with its current position.
           */
          [NSApp sendEvent: theEvent];
          break;
          
        case GSAppKitDraggingStatus:
          NSDebugLLog(@"NSDragging", @"got GSAppKitDraggingStatus\n");
          if ([theEvent data1] == targetWindow)
            {
              targetMask = [theEvent data2];

              [self _setCursor];
            }
          break;
          
        case GSAppKitDraggingFinished:
          NSLog(@"Internal: got GSAppKitDraggingFinished out of seq");
          break;
          
        case GSAppKitWindowFocusIn:
          break;
          
        default:
          NSLog(@"Internal: dropped NSAppKitDefined (%d) event", sub);
          break;
        }
      }
      break;
      
    case NSMouseMoved:
    case NSLeftMouseDragged:
    case NSLeftMouseDown:
    case NSLeftMouseUp:
      newPosition = [[theEvent window] convertBaseToScreen: [theEvent locationInWindow]];
      break;
    case NSFlagsChanged:
      if ([self _updateOperationMask: theEvent])
        {
          [self _setCursor];
        }
      break;
    case NSPeriodic:
      // targetWindow check is needed because otherwise events only arrive after an initial
      // mouse move.
      if (newPosition.x != dragPosition.x || newPosition.y != dragPosition.y) 
        {
          [self _updateAndMoveImageToCorrectPosition];
        }
      break;
    default:
      NSLog(@"Internal: dropped event (%d) during dragging", [theEvent type]);
    }
}
  
/*
 * This method will move the drag image, update and update all associated data
 */

- (void) _updateAndMoveImageToCorrectPosition
{
  NSWindow		*oldDragWindow;
  BOOL                   oldDragExternal;
  gswindow_device_t	*dwindev;
  Window                 mouseWindow; 
  BOOL                   changeCursor = NO;
            
  //--- Move drag image to the new position -----------------------------------
            
  /* Move drag window, because X coordinates are flipped the wy and wx calculation are different */ 
  wx += (int) (newPosition.x - dragPosition.x);
  wy += (int) (dragPosition.y - newPosition.y);
            
  XMoveWindow (XDPY, dragWindev->ident, wx, wy);
            
  //--- Store old values -----------------------------------------------------
            
  dragPosition = newPosition;
  oldDragWindow = dragWindow;
  oldDragExternal = dragExternal;
            
            
  //--- Determine target XWindow ---------------------------------------------
            
  mouseWindow = [self _xWindowAcceptingDnDunderX: wx + offset.x Y: wy + offset.y];

  //--- Determine target NSWindow --------------------------------------------

  dwindev = [XGContext _windowForXWindow: mouseWindow];
            
  if (dwindev != 0)
    {
      dragWindow = GSWindowWithNumber(dwindev->number);
    }
  else
    {
      dragWindow = nil;
    }

  // If we have are not hovering above a window that we controll
  // we are dragging to an external application.
            
  dragExternal = (mouseWindow != (Window) None) && (dragWindow == nil);
            
  if (dragWindow)
    {
      dragPoint = [dragWindow convertScreenToBase: dragPosition];
    }
            
  NSDebugLLog(@"NSDragging", @"mouse window %x\n", mouseWindow);
            
            
            
  //--- send exit message if necessary -------------------------------------
            
  if ((mouseWindow != targetWindow) && targetWindow)
    {
      /* If we change windows and the old window is dnd aware, we send an
         dnd exit */
                
      NSDebugLLog(@"NSDragging", @"sending dnd exit\n");
                
      if (oldDragWindow)   
        {
          [self _sendLocalEvent: GSAppKitDraggingExit
                action: dragMask & operationMask
                position: NSZeroPoint
                timestamp: dragSequence
                toWindow: oldDragWindow];
        }  
      else
        {  
          xdnd_send_leave(&dnd, targetWindow, dragWindev->ident);
        }
    }

  //  Reset drag mask when we switch from external to internal or back
  //
  if (oldDragExternal != dragExternal)
    {
      int newMask = [dragSource draggingSourceOperationMaskForLocal: dragExternal];

      if (newMask != dragMask)
        {
          dragMask = newMask;
          changeCursor = YES;
        }
    }


  if (mouseWindow == targetWindow && targetWindow)  
    { // same window, sending update
      NSDebugLLog(@"NSDragging", @"sending dnd pos\n");
      if (dragWindow)
        {
          [self _sendLocalEvent: GSAppKitDraggingUpdate
                action: dragMask & operationMask
                position: NSMakePoint (wx + offset.x, wy + offset.y)
                timestamp: CurrentTime
                toWindow: dragWindow];
        }
      else
        {
          xdnd_send_position(&dnd,
                             targetWindow, dragWindev->ident,
                             GSActionForDragOperation (dragMask & operationMask), wx + offset.x, 
                             wy + offset.y, CurrentTime);
        }
    }
  else if (mouseWindow != (Window) None)
    {
      //FIXME: We might force the cursor update here, if the
      //target wants to change the cursor.
      
      NSDebugLLog(@"NSDragging",
                  @"sending dnd enter/pos\n");
      
      if (dragWindow)
        {
          [self _sendLocalEvent: GSAppKitDraggingEnter
                action: dragMask
                position: NSMakePoint (wx + offset.x, wy + offset.y)
                timestamp: CurrentTime
                toWindow: dragWindow];
        }
      else
        {
          xdnd_send_enter(&dnd,
                          mouseWindow, dragWindev->ident, typelist);
          xdnd_send_position(&dnd,
                             mouseWindow, dragWindev->ident,
                             GSActionForDragOperation (dragMask & operationMask),
                             wx + offset.x, wy + offset.y, CurrentTime);
        }
    }

  if (targetWindow != mouseWindow)
    {
      targetWindow = mouseWindow;
      changeCursor = YES;
    }
  
  if (changeCursor)
    {
      [self _setCursor];
    }
}

/* NSDraggingInfo protocol */
- (NSWindow*) draggingDestinationWindow
{
  return dragWindow;
}

- (NSPoint) draggingLocation
{
  return dragPoint;
}

- (NSPasteboard*) draggingPasteboard
{
  return dragPasteboard;
}

- (int) draggingSequenceNumber
{
  return dragSequence;
}

- (id) draggingSource
{
  return dragSource;
}

- (unsigned int) draggingSourceOperationMask
{
  return dragMask;
}

- (NSImage*) draggedImage
{
  if (dragSource)
    return [dragCell image];
  else
    return nil;
}

- (NSPoint) draggedImageLocation
{
  NSPoint loc;

  if (dragSource)
    {
      NSSize size;
      size = [[dragCell image] size];
      loc = NSMakePoint(dragPoint.x-size.width/2, dragPoint.y - size.height/2);
    }
  else
    {
      loc = dragPoint;
    }
  return loc;
}

- (void) slideDraggedImageTo: (NSPoint)screenPoint
{
  // FIXME
}


/*
  return search all descendents of parent and return
  window under mouse that accepts drag and drop.
  -1     if we can only find the window we are dragging under the mouse.
  None   if there is no window under the mouse that accepts drag and drop.
*/
- (Window) _xWindowAcceptingDnDDescendentOf: (Window) parent underX: (int) x Y: (int) y
{
  Window *children;
  int nchildren;
  Window  result = None;
  Window  ignore, child2, root;
  Display *display = XDPY;
  XWindowAttributes attr;
  int ret_x, ret_y;

  if (parent == dragWindev->ident)
    return -1;
  
  XQueryTree(display, parent, &root, &ignore, &children, &nchildren);

  while (nchildren-- > 0)
    {
      Window child = children [nchildren];
  
      if (XGetWindowAttributes (display, child, &attr)
          && attr.map_state == IsViewable 
          && XTranslateCoordinates (display, root, child, x, y, &ret_x, &ret_y, &child2)
          && ret_x >= 0 && ret_x < attr.width
          && ret_y >= 0 && ret_y < attr.height)
        {
          result = [self _xWindowAcceptingDnDDescendentOf: child underX: x Y: y];
          if (result != -1)
            break;
        }
    }

  if (children)
      XFree (children);
  
  if (result == (Window) None)
    {
      if (xdnd_is_dnd_aware (&dnd, parent, &dnd.dragging_version, typelist))
        {
          result = parent;
        }
    }
  
  return result;
}

/*
  Return window under the mouse that accepts drag and drop
 */
- (Window) _xWindowAcceptingDnDunderX: (int) x Y: (int) y
{
  Window result = [self _xWindowAcceptingDnDDescendentOf: dragWindev->root underX: x Y: y];

  if (result == -1)
    return None;
  else
    return result;
}

@end



@interface XGContext (DragAndDrop)

- (void) _resetDragTypesForWindow: (NSWindow *)win;
- (BOOL) _addDragTypes: (NSArray*)types toWindow: (NSWindow *)win;
- (BOOL) _removeDragTypes: (NSArray*)types fromWindow: (NSWindow *)win;
- (void) _postExternalEvent: (NSEvent *)theEvent;

@end


@implementation XGContext (DragAndDrop)

- (void) _resetDragTypesForWindow: (NSWindow *)win
{
  int			winNum;
  Atom			*typelist;
  gswindow_device_t	*window;
  NSCountedSet		*drag_set = [self _dragTypesForWindow: win];

  winNum = [win windowNumber];
  window = [XGContext _windowWithTag: winNum];

  GSEnsureDndIsInitialized ();

  typelist = mimeTypeForPasteboardType(XDPY, [self zone],[drag_set allObjects]);
  NSDebugLLog(@"NSDragging", @"Set types on %x to %@", window->ident, drag_set);
  xdnd_set_dnd_aware(&dnd, window->ident, typelist);

  NSZoneFree([self zone], typelist);
}

- (BOOL) _addDragTypes: (NSArray*)types toWindow: (NSWindow *)win
{
  BOOL	did_add;
  int	winNum;

  did_add = [super _addDragTypes: types toWindow: win];
  /* Check if window device exists */
  winNum = [win windowNumber];
  if (winNum > 0 && did_add == YES)
    {
      [self _resetDragTypesForWindow: win];
    }
  return did_add;
}

- (BOOL) _removeDragTypes: (NSArray*)types fromWindow: (NSWindow *)win
{
  BOOL	did_change;
  int	winNum;

  did_change = [super _removeDragTypes: types fromWindow: win];
  /* Check if window device exists. */
  winNum = [win windowNumber];
  if (winNum > 0 && did_change == YES)
    {
      [self _resetDragTypesForWindow: win];
    }
  return did_change;
}

- (void) _postExternalEvent: (NSEvent *)theEvent
{
  gswindow_device_t	*window;

  window = [XGContext _windowWithTag: [[theEvent window] windowNumber]];
  if ([theEvent subtype] == GSAppKitDraggingStatus)
    {
      Atom		xaction;
      NSDragOperation	action = [theEvent data2];

      xaction = GSActionForDragOperation(action);
      xdnd_send_status(&dnd, 
		       [theEvent data1],
		       window->ident,
		       (action != NSDragOperationNone),
		       0,
		       0, 0, 0, 0,
		       xaction);
    }
  else if ([theEvent subtype] == GSAppKitDraggingFinished)
    {
      xdnd_send_finished(&dnd, 
			 [theEvent data1],
			 window->ident,
			 0);
    }
  else
    {
      NSDebugLLog(@"NSDragging", @"Internal: unhandled post external event");
    }
}

@end




@implementation XGRawWindow

- (BOOL) canBecomeMainWindow
{
  return NO;
}

- (BOOL) canBecomeKeyWindow
{
  return NO;
}

- (void) _initDefaults
{
  [super _initDefaults];
  [self setReleasedWhenClosed: NO];
  [self setExcludedFromWindowsMenu: YES];
}

- (id) initWithContentRect: (NSRect)contentRect
		 styleMask: (unsigned int)aStyle
		   backing: (NSBackingStoreType)bufferingType
		     defer: (BOOL)flag
		    screen: (NSScreen*)aScreen
{
  self = [super initWithContentRect: contentRect
			  styleMask: aStyle
			    backing: bufferingType
			      defer: flag
			     screen: aScreen];

  [[NSApplication sharedApplication] removeWindowsItem: self];

  return self;
}

- (void) orderWindow: (NSWindowOrderingMode)place relativeTo: (int)otherWin
{
  XSetWindowAttributes winattrs;
  unsigned long valuemask;
  gswindow_device_t *window;

  [super orderWindow: place relativeTo: otherWin];

  window = [XGContext _windowWithTag: _windowNum];
  valuemask = (CWSaveUnder|CWOverrideRedirect);
  winattrs.save_under = True;
  /* Temporarily make this False? we don't handle it correctly (fedor) */
  winattrs.override_redirect = False;
  XChangeWindowAttributes (XDPY, window->ident,
                           valuemask, &winattrs);
}

@end
